/////////////////////////////////////////////////////////////////////////////
// Invaders.cpp
//
// Main code for invaders. Uses Os independency layer only. Includes local
// include files only.
/////////////////////////////////////////////////////////////////////////////

#include "Invaders.h"

// Index of left, right, and bottom most columns with invaders

int gnLeftColumn;
int gnTopRow;
int gnRightColumn;
int gnBottomRow;

// Index of row currently moving

int gnRowMove;

// left, top, right, edges of invader field

int gxLeftInvader;
int gxRightInvader;
int gyTopInvader;
int gyTopInvaderStart;

// Index of bottom most invader row for each column

char ganRowLastInvader[kcInvaderColumns];

// # of non-empty invader columns

int gcNonEmptyColumns;

// # of invaders left

int gcInvaders;

// Byte array, set to non-zero if an invader is present

byte gabFInvader[kcInvaderColumns][kcInvaderRows];

// Direction and amount invaders are moving in

int gdxInvaderMove;

// x of ship

int gxShip;

// Accumulates time. When over kmsShipMove, ship can move

int gmsShipMoveAcc;

// Total amount of time to move invaders

int gmsMoveTotal;

// # ms between invader row moves

int gmsRowMove;

// Accumulates time. When over gmsRowMove, row can move

int gmsRowMoveAcc;

// Amount to speedup invaders for each kill

int gmsSpeedup;

// Used to time invader fire so that an invader doesn't fire
// on top of a just fired missle

int gmsInvaderFireAcc;

// Which set of invaders we're using right now

int gnInvaderSelect;

// # of ships left

int gcShips;

// # of frames for ship explosion

int gcShipExplosionFrames;

// Time accumulator for timing ship explosion

int gmsShipExplosionAcc;

// Which of 4 walk sounds to play

int gnWalkSound;

// Time between walk sounds, plus walk sound accumulator

int gmsWalkSoundAcc;
int gmsWalkSound;

// Game state

int gnState;

// Demo state

int gnDemoState;

// First missle is ship missle, rest are invaders

int const kcMissles = 1 + kcInvaderMissles;
Missle gamsl[kcMissles];
#define gmslShip gamsl[0]
#define gamslInvaders (&gamsl[1])

// 2 stances of invaders:
//
// Invader stance0:
// top open
// middle open
// middle open
// bottom closed
// bottom closed
//
// Invader stance1:
// top closed
// middle closed
// middle closed
// bottom open
// bottom open

byte gmpInvaderBitmap[2][5] = {
	(byte)kbidTopInvaderOpen,
	(byte)kbidMiddleInvaderOpen,
	(byte)kbidMiddleInvaderOpen,
	(byte)kbidBottomInvaderClosed,
	(byte)kbidBottomInvaderClosed,
	(byte)kbidTopInvaderClosed,
	(byte)kbidMiddleInvaderClosed,
	(byte)kbidMiddleInvaderClosed,
	(byte)kbidBottomInvaderOpen,
	(byte)kbidBottomInvaderOpen
};

// Mothership globals

int gxMotherShip;
int gdxMotherShipMove;
int gmsMotherShipMoveAcc;
int gmsMotherShipWaitAcc;

// Score
// LATER: deal with top score of 65535, make 99999.

int const gnIndexScore = 4;
int const gnIndexHighScore = 4;
TCHAR gszScore[] = TEXT("00000");
TCHAR gszHighScore[] = TEXT("00000");
bool gfUpdateScore;
unsigned int gunScore;
unsigned int gunHighScore;
unsigned int kunMaxScore = 65535;

// Score for each row

byte ganInvaderScore[] = {
	30,
	20,
	20,
	10,
	10
};

// Demo help screen time accumulator

int gmsDemoHelpAcc;
void InsertDecimalAscii(TCHAR *psz, short s, short c)
{
	while (c-- != 0) {
		*psz-- = (OsMod(s, 10)) + '0';
		s = OsDiv(s, 10);
	}
}

void AddScore(int n)
{
	// If we're in demo mode, don't modify score

	if (gnState == knDemoMode)
		return;

	// If at max score, peg there

	if (gunScore > kunMaxScore - n) {
		gunScore = kunMaxScore;
	} else {
		gunScore += n;
	}
	InsertDecimalAscii(&gszScore[gnIndexScore], gunScore, 5);

	if (gunHighScore < gunScore) {
		gunHighScore = gunScore;
		InsertDecimalAscii(&gszHighScore[gnIndexHighScore], gunHighScore, 5);
	}

	// Redraw score next time through

	gfUpdateScore = TRUE;
}

bool IsInvaderPresent(int i, int j)
{
	return gabFInvader[i][j] != 0;
}

void ClearInvader(int i, int j)
{
	gabFInvader[i][j] = 0;
	gcInvaders--;

	// Add to score

	AddScore((int)ganInvaderScore[j]);

	// What is the last invader in this column?

	for (int k = kcInvaderRows - 1; k >= 0; k--) {
		if (IsInvaderPresent(i, k)) {
			ganRowLastInvader[i] = (char)k;
			return;
		}
	}

	// No more invaders in this column!

	gcNonEmptyColumns--;
	ganRowLastInvader[i] = -1;
}

void DrawRow(int nRow, int x, int y)
{
	int bid = (int)(gmpInvaderBitmap[gnInvaderSelect][nRow]);

	int xT = x;
	for (int i = gnLeftColumn; i <= gnRightColumn; i++) {
		if (IsInvaderPresent(i, nRow))
			OsBlt(bid, xT, y);
		xT += kcxInvader;
	}
}

void GameStart()
{
	// Game start sound

	if (gnState != knDemoMode)
		OsPlaySound(kSndGameStart);

	// Y that invaders start at

	gyTopInvaderStart = kyInvaderLevel1;

	// Next state is starting the level

	gnState = knLevelStart;

	// # of ships to start with

	gcShips = kcShipsStart;
}

void LevelStart()
{
	// Fill screen w/black

	OsFill(kxScreenLeft, kyScreenTop, kcxScreen, kcyScreen, FALSE);

	// Turn on all invaders

	memset(gabFInvader, 0xff, sizeof(gabFInvader));

	// Initialize invader globals

	gnLeftColumn = 0;
	gnRightColumn = kcInvaderColumns - 1;
	gnBottomRow = kcInvaderRows - 1;
	gxLeftInvader = kxInvaderStart;
	gxRightInvader = kxInvaderStart + kcxInvader * kcInvaderColumns;
	gyTopInvader = gyTopInvaderStart;
	gdxInvaderMove = kcxInvaderMove;

	// This is post incremented, but never progresses further than the max
	// (which puts the bottom invader right above the ship!)

	gyTopInvaderStart += kcyInvaderMove + kcyInvaderInvaderSpace;
	if (gyTopInvaderStart > kyInvaderStartMax)
		gyTopInvaderStart = kyInvaderStartMax;

	// Index of left, right, and bottom most columns with invaders

	gnLeftColumn = 0;
	gnTopRow = 0;
	gnRightColumn = kcInvaderColumns - 1;
	gnBottomRow = kcInvaderRows - 1;

	// Index of row currently moving

	gnRowMove = gnBottomRow;

	// All invaders present in each row

	memset(ganRowLastInvader, gnBottomRow, sizeof(ganRowLastInvader));
	gcNonEmptyColumns = kcInvaderColumns;

	// Time between invader moves

	gmsMoveTotal = kmsMoveTotal;
	gmsRowMove = OsDiv(kmsMoveTotal, kcInvaderRows);

	// Time to speedup invaders after each kill
	
	gmsSpeedup = kmsSpeedupInit;

	// Draw the platform

	OsFill(kxPlatform, kyPlatform, kcxPlatform, kcyPlatform, TRUE);

	// Draw the shields

	int x = kxShieldFirst;
	for (int i = 0; i < kcShields; i++) {
		OsBlt(kbidShield, x, kyShieldFirst);
		x += gabrc[kbidShield].cx + kcxShieldShieldSpace;
	}

	// Ship starts at left edge

	gxShip = kxLeftShipBoundary;
	OsBlt(kbidShip, gxShip, kyShip);

	// Draw invaders - select invader set 0

	gnInvaderSelect = 0;
	int y = gyTopInvader + (kcInvaderRows - 1) * (kcyInvader + kcyInvaderInvaderSpace);

	// Invaders don't fire until all rows drawn
	
	gmsInvaderFireAcc = -(gmsRowMove * kcInvaderRows);
	gnInvaderSelect = 1;

	// No missles firing

	memset(&gamsl, 0, sizeof(gamsl));

	// # of invaders left

	gcInvaders = kcInvaderRows * kcInvaderColumns;

	// Reset time waiting for mothership

	gmsMotherShipWaitAcc = 0;
	gxMotherShip = -1;

	// Next state is playing the game

	gnState = knGamePlay;

	// Update score

	gfUpdateScore = TRUE;

	// Turn all missles off

	for (i = 0; i < kcMissles; i++)
		gamsl[i].nState = kMslNone;

	// Not exploding

	gcShipExplosionFrames = 0;

	// Walk stance 0

	gnWalkSound = 0;
	gmsWalkSoundAcc = 0;
//	gmsWalkSound = kmsWalkSoundInit;
	gmsWalkSound = gmsMoveTotal;
}

void MoveShip()
{
	if (!(gfusInput & (kfLeftKey | kfRightKey)))
		return;

	gmsShipMoveAcc += kmsOnTimer;
	while (gmsShipMoveAcc >= kmsShipMove) {
		gmsShipMoveAcc -= kmsShipMove;
		if (gfusInput & kfLeftKey) {
			gxShip -= kdxShipMove;
			if (gxShip < kxLeftShipBoundary)
				gxShip = kxLeftShipBoundary;
		} else if (gfusInput & kfRightKey) {
			gxShip += kdxShipMove;
			if (gxShip > kxRightShipBoundary)
				gxShip = kxRightShipBoundary;
		}
		OsBlt(kbidShip, gxShip, kyShip);
	}
}

void ExplodeShip(bool fInit)
{
	// Initialize explosion?

	if (fInit) {
		gmsShipExplosionAcc = kmsShipExplosion;
		gcShipExplosionFrames = kcShipExplosionFrames;
	}

	// If ship not exploding, get out of here

	if (gcShipExplosionFrames == 0)
		return;

	// Deal with timing

	gmsShipExplosionAcc += kmsOnTimer;
	if (gmsShipExplosionAcc < kmsShipExplosion)
		return;
	gmsShipExplosionAcc -= kmsShipExplosion;

	// Figure out explosion frame

	int bid = kbidShipExplosionFrame1 + (gcShipExplosionFrames & 1);
	OsBlt(bid, gxShip, kyShip + kcyShip - kcyShipExplosion);
	gcShipExplosionFrames--;

	// Play ship explosion sound

	if (fInit)
		OsPlaySound(kSndBaseHit);

	// If we're out of frames, grab the next ship

	if (gcShipExplosionFrames == 0) {
		// Erase where the explosion was
		
		OsFill(gxShip, kyShip + kcyShip - kcyShipExplosion, kcxShipExplosion,
				kcyShipExplosion, FALSE);
		gfUpdateScore = TRUE;

		// Any more ships? If not, game over. If yes, update
		// score line which will update # of ships left

		gcShips--;
		if (gcShips == 0) {
			gnState = knGameEnd;
			return;
		}

		// If mothership is still moving, resume mothership sound

		if (gxMotherShip != -1)
			OsPlaySound(kSndMotherWalk);

		// Start ship on left edge

		gxShip = kxLeftShipBoundary;

		// Redraw the image

		OsBlt(kbidShip, gxShip, kyShip);

		// Delay the next invader missle by 1/2 second

		gmsInvaderFireAcc = -500;
	}
}

void PlayWalkSound()
{
	// Don't play walk sound if ship missle exploding invader

	if (gmslShip.nState == kMslExploding && gmslShip.bid == kbidInvaderExplosion)
		return;

	gmsWalkSoundAcc += kmsOnTimer;
	while (gmsWalkSoundAcc >= gmsWalkSound) {
		gmsWalkSoundAcc -= gmsWalkSound;

		// Play walk sound, advance to next walk sound

		OsPlaySound(kSndInvaderWalk0 + gnWalkSound);
		gnWalkSound++;
		if (gnWalkSound > 3)
			gnWalkSound = 0;
	}
}

void MoveInvaders()
{
	// Don't move invaders if ship missle exploding invader

	if (gmslShip.nState == kMslExploding && gmslShip.bid == kbidInvaderExplosion)
		return;

	gmsRowMoveAcc += kmsOnTimer;
	while (gmsRowMoveAcc >= gmsRowMove) {
		gmsRowMoveAcc -= gmsRowMove;

		int xLeft = gxLeftInvader + gdxInvaderMove;
		int xRight = gxRightInvader + gdxInvaderMove;
		int y = gyTopInvader + gnRowMove * (kcyInvader + kcyInvaderInvaderSpace);
		if (xLeft < kxLeftInvaderBoundary || xRight > kxRightInvaderBoundary) {
			DrawRow(gnRowMove, gxLeftInvader, y + kcyInvaderMove);
			OsFill(gxLeftInvader, y, gxRightInvader - gxLeftInvader,
					kcyInvaderMove, FALSE);

			// If we've hit the bottom, it's game over

			if (y + kcyInvaderMove > kyBottomInvaderBoundary) {
				gcShips = 1;
				ExplodeShip(TRUE);
				return;
			}

			// Haven't hit bottom, move the next row down. If we've moved all rows,
			// it's time to move in the opposite direction.

			gnRowMove--;
			if (gnRowMove < gnTopRow) {
				gnRowMove = gnBottomRow;
				gyTopInvader += kcyInvaderMove;
				gdxInvaderMove *= -1;
				gnInvaderSelect ^= 1;
				gfUpdateScore = TRUE;
			}
			return;
		}

		// Not time to move down - we're still within boundaries. Move this row.	

		DrawRow(gnRowMove, xLeft, y);

		// Haven't hit bottom, move the next row down. If we've moved all rows,
		// it's time to move in the opposite direction.

		// Move next row. If we're at the top (last row to move) then adjust the
		// overall x position!

		gnRowMove--;
		if (gnRowMove < gnTopRow) {
			gnRowMove = gnBottomRow;
			gxLeftInvader = xLeft;
			gxRightInvader = xRight;
			gnInvaderSelect ^= 1;
		}
	}
}

void FireShipMissle()
{
	if (gmslShip.nState != kMslNone)
		return;
	if (!(gfusInput & kfFireKey))
		return;

	memset(&gmslShip, 0, sizeof(Missle));
	gmslShip.nState = kMslAnimating;
	gmslShip.x = gxShip + kcxShip / 2 - ((kcxShip & 1) ^ 1);
	gmslShip.y = kyShip - kcyShipMissle;
	gmslShip.dy = -kdyShipMissleMove;
	gmslShip.msMove = kmsShipMissleMove;
	gmslShip.msMoveAcc = kmsShipMissleMove;
	gmslShip.cBitmaps = 1;
	gmslShip.cBitmapsTrail = 5;
	gmslShip.bid = kbidShipMissle;

	OsPlaySound(kSndFire);
}

void CalcInvaderPos(int nCol, int nRow, int *px, int *py)
{
	// Calc x, y of invader position

	int i = nCol - gnLeftColumn;
	int j = nRow - gnTopRow;

	*px = gxLeftInvader + i * (kcxInvader + kcxInvaderInvaderSpace);
	*py = gyTopInvader + j * (kcyInvader + kcyInvaderInvaderSpace);

	// Make x or y adjustment for current invader positions

	if (nRow > gnRowMove) {
		if ((gdxInvaderMove < 0 && gxLeftInvader <= kxLeftInvaderBoundary) ||
				(gdxInvaderMove > 0 && gxRightInvader >= kxRightInvaderBoundary)) {
			*py += kcyInvaderMove;
		} else {
			*px += gdxInvaderMove;
		}
	}
}

void ExplodeMotherShip(bool fInit)
{
	static short msMotherShipExplosionAcc;
	static byte mpScoreIndex[] = {
		0, 0, 0, 0, 0,
		1, 1,
		2
	};
	static short asScore[] = {
		50, 150, 300
	};
	
	if (fInit) {
		// Turn this on

		msMotherShipExplosionAcc = 1;

		// Erase mothership

		OsFill(kxScreenLeft, kyMotherShip, kcxScreen, kcyMotherShip, FALSE);

		// Calc score text clumsily. Avoids fixups that a string array causes
		// (Pilot loader right now can't support that)
		
		int n = mpScoreIndex[OsRand() & 7];

		// Draw score in appropriate position

		OsBlt(kbid50Pts + n, gxMotherShip + (kcxMotherShip - gabrc[n + kbid50Pts].cx) / 2,
				kyMotherShip);
		AddScore(asScore[n]);

		// Celebrate
		
		OsPlaySound(kSndMotherHit);

		// Turn off MotherShip
		
		gxMotherShip = -1;
		return;
	}

	// If we've timed out, get rid of score

	if (msMotherShipExplosionAcc == 0)
		return;
	msMotherShipExplosionAcc += kmsOnTimer;
	if (msMotherShipExplosionAcc < kmsMotherShipExplosion)
		return;

	// Get rid of score, turn all this off

	msMotherShipExplosionAcc = 0;
	OsFill(kxScreenLeft, kyMotherShip, kcxScreen, kcyMotherShip, FALSE);
}

void ExplodeMissle(Missle *pmsl, bool fOffScreen)
{
	// The ship missle is missle 0

	if (pmsl == &gmslShip) {
		// See if it hit an invader.

		int nCol = -1;
		int nRow = -1;
		if (pmsl->x >= gxLeftInvader && pmsl->y >= gyTopInvader) {
			int i = OsDiv(pmsl->x - gxLeftInvader, kcxInvader + kcxInvaderInvaderSpace);
			int j = OsDiv(pmsl->y - gyTopInvader, kcyInvader + kcyInvaderInvaderSpace);
			nCol = i + gnLeftColumn;
			nRow = j + gnTopRow;
		}

		// Invader here?

		if (!fOffScreen && nCol >= 0 && nCol < kcInvaderColumns &&
				nRow >= 0 && nRow < kcInvaderRows &&
				IsInvaderPresent(nCol, nRow)) {

			// Invader is gone! Clear invader.

			ClearInvader(nCol, nRow);

			// Get invader x & y

			CalcInvaderPos(nCol, nRow, &pmsl->x, &pmsl->y);
			
			// Draw the invader explosion here
			
			pmsl->bid = kbidInvaderExplosion;
			OsBlt(kbidInvaderExplosion, pmsl->x, pmsl->y);

			// Sound effect

			OsPlaySound(kSndInvaderHit);
		} else {
			// Not an invader, but a regular explosion.
			
			if (!fOffScreen) {
				if (pmsl->y < kyMotherShip) {
					// Missle hit score line. Let it continue.

					return;
				} else if (pmsl->y < kyMotherShip + kcyMotherShip) {
					// May be the mothership, or score from mother ship

					if (pmsl->x >= gxMotherShip && (pmsl->x + kcxInvaderMissle) <
							(gxMotherShip + kcxMotherShip)) {
						ExplodeMotherShip(TRUE);
						pmsl->nState = kMslNone;
						return;
					}

					// hit mothership score, let missle sail through

					return;
				}
			}

			// Either off-screen or hit a shield. Calc x & y, bound by
			// screen, and explode. Should take 8 shots to shoot through the
			// thick part of a shield.

			pmsl->x -= (kcxShipMissleExplosion - kcxShipMissle) / 2;
			pmsl->y -= kdxShipMissleExplosionAdjust;
 			if (pmsl->y < kyScreenTop)
				pmsl->y = kyScreenTop;
			pmsl->bid = kbidBulletExplosion;
			OsOrBlt(kbidBulletExplosion, pmsl->x, pmsl->y);
		}
	} else {
		// Invader missle hit something! Is it the ship?

		if (pmsl->y + kcyInvaderMissle >= kyShip && pmsl->y + kcyInvaderMissle < kyShip + kcyShip) {
			// It might be another missle explosion that got hit, and
			// not the ship. Check for bounds of ship first!

			if (pmsl->x + kcxInvaderMissle >= gxShip && pmsl->x < gxShip + kcxShip) {
				ExplodeShip(TRUE);
				pmsl->nState = kMslNone;
				return;
			}

			// Continue advancing so we don't explode on top of explosions

			return;
		}

		// Project the explosion forward so the top of
		// it is in line with the tip of the missle. Explosions
		// on a fresh shield should be aligned with the top of the shield.

		pmsl->x -= (kcxInvaderMissleExplosion - kcxInvaderMissle) / 2;
		if (pmsl->y + kcyInvaderMissleExplosion >= kyScreenBottom)
			pmsl->y = kyScreenBottom - kcyInvaderMissleExplosion;
		pmsl->bid = kbidMissleExplosion;
		OsOrBlt(kbidMissleExplosion, pmsl->x, pmsl->y);
	}

	// Common state for all explosions

	pmsl->nState = kMslExploding;	
	pmsl->msMove = kmsExplosionTimeout;
	pmsl->msMoveAcc = 0;
}

void ClearExplosion(Missle *pmsl)
{
	if (pmsl->bid == kbidInvaderExplosion) {
		// Hack: we reused the mask bits for something else so just clear it

		OsFill(pmsl->x, pmsl->y, gabrc[kbidInvaderExplosion].cx, gabrc[kbidInvaderExplosion].cy, FALSE);

	} else {
		// Erase missle and turn it off

		OsNotAndBlt(pmsl->bid, pmsl->x, pmsl->y);
	}
	pmsl->nState = kMslNone;

	// If missle from ship...

	if (pmsl == &gmslShip) {
		// Only fire again once the fire key goes up if this is the ship
		// missle exploding
		
		if (gfusInput & kfFireKey) {
			gfusInput &= ~kfFireKey;
			gfusInput |= kfWaitFireKeyUp;
		}

		// If explosion at top of the screen, update score so it
		// doest have a hole in it from the explosion

		if (pmsl->y < kyScore + kcyScore)
			gfUpdateScore = TRUE;
	}

	// If this missle was an invader exploding, recheck boundries
	// to define new edges.

	if (pmsl->bid != kbidInvaderExplosion)
		return;

	// Look along bottom edge

	bool fFound = FALSE;
	for (int i = gnBottomRow; i >= gnTopRow; i--) {
		for (int j = gnLeftColumn; j <= gnRightColumn; j++) {
			if (IsInvaderPresent(j, i)) {
				fFound = TRUE;
				break;
			}
		}
		if (fFound)
			break;
		gnBottomRow--;
	}

	// Look along left edge

	fFound = FALSE;
	for (i = gnLeftColumn; i <= gnRightColumn; i++) {
		for (int j = gnTopRow; j <= gnBottomRow; j++) {
			if (IsInvaderPresent(i, j)) {
				fFound = TRUE;
				break;
			}
		}
		if (fFound)
			break;
		gnLeftColumn++;
		gxLeftInvader += kcxInvader + kcxInvaderInvaderSpace;
	}

	// Look along right edge

	fFound = FALSE;
	for (i = gnRightColumn; i >= gnLeftColumn; i--) {
		for (int j = gnTopRow; j <= gnBottomRow; j++) {
			if (IsInvaderPresent(i, j)) {
				fFound = TRUE;
				break;
			}
		}
		if (fFound)
			break;
		gnRightColumn--;
		gxRightInvader -= kcxInvader + kcxInvaderInvaderSpace;
	}

	// Adjust time to move invaders.

	if (gcInvaders < kcInvadersSpeedupAfter) {
		gmsMoveTotal -= gmsSpeedup;
		gmsSpeedup += kmsSpeedupInc;
	}
	int cRows = gnBottomRow - gnTopRow + 1;
	if (cRows != 0)
		gmsRowMove = OsDiv(gmsMoveTotal, cRows);

	// Adjust time between walk sounds

//	gmsWalkSound -= kmsWalkSoundSpeedup;
	gmsWalkSound = gmsMoveTotal / 2;
}

void MoveMissles()
{
	for (int i = 0; i < kcMissles; i++) {
		// Missle in this slot?

		Missle *pmsl = &gamsl[i];
		if (pmsl->nState == kMslNone)
			continue;

		// See if it's time to do something

		pmsl->msMoveAcc += kmsOnTimer;
		while (pmsl->msMoveAcc >= pmsl->msMove) {
			pmsl->msMoveAcc -= pmsl->msMove;

			// If exploding, we're done with explosion. Erase it and
			// turn off missle.
			
			if (pmsl->nState == kMslExploding) {
				ClearExplosion(pmsl);
				break;
			}

			// Figure out the next image frame, new position. If past
			// screen edge, automatic explosion.

			pmsl->nBitmap++;
			int nBitmap = OsMod(pmsl->nBitmap, pmsl->cBitmaps);
			int nBitmapEnd = OsMod(pmsl->nBitmap - pmsl->cBitmapsErase, pmsl->cBitmaps);
			pmsl->y += pmsl->dy;
			int yEnd = pmsl->y + (-pmsl->dy * pmsl->cBitmapsErase);
			bool fOffScreen;
			int bid = nBitmap + pmsl->bid;
			if (OsCheckBits(pmsl->x, pmsl->y, gabrc[bid].cx, gabrc[bid].cy, &fOffScreen)) {
				// Erase missle

				for (int j = 0; j < pmsl->cBitmapsErase; j++) {
					OsNotAndBlt(nBitmapEnd + pmsl->bid, pmsl->x, yEnd);
					nBitmapEnd = OsMod(nBitmapEnd + 1, pmsl->cBitmaps);
					yEnd += pmsl->dy;
				}

				// setup explosion here

				ExplodeMissle(pmsl, fOffScreen);
				break;
			}

			// No collision. OR in new part of missle, AND old. Old image
			// is the same as the one we're OR-ing in.

			OsOrBlt(bid, pmsl->x, pmsl->y);
			if (pmsl->cBitmapsErase == pmsl->cBitmapsTrail)
				OsNotAndBlt(nBitmapEnd + pmsl->bid, pmsl->x, yEnd);

			// This is so we don't start erasing multi-segmented missles
			// until they are full length.

			pmsl->cBitmapsErase++;
			if (pmsl->cBitmapsErase > pmsl->cBitmapsTrail)
				pmsl->cBitmapsErase = pmsl->cBitmapsTrail;
		}
	}
}

void FireInvaderMissles()
{
	// If no invaders left, return

	if (gcNonEmptyColumns == 0)
		return;

	// This missle timing is only here to make sure that one missle doesn't
	// get launched on top of another, so it checks for 8 missle animations
	// between shot firings

	if (gmsInvaderFireAcc < 8 * kmsInvaderMissleMove) {
		gmsInvaderFireAcc += kmsOnTimer;
		return;
	}

	// See if it is time to fire a missle
	// This'll fire 1 missle every kcTicksPerInvaderFire ticks.

	if (OsMod(OsRand(), kcTicksPerInvaderFire) != 0)
		return;

	// If no missles available, get out.

	int nMslReady;
	for (nMslReady = 1; nMslReady < kcMissles; nMslReady++) {
		if (gamsl[nMslReady].nState == kMslNone)
			break;
	}
	if (nMslReady == kcMissles)
		return;

	// We're firing a missle. Reset the invader fire timer
	
	gmsInvaderFireAcc = 0;

	// Pick one of the invader columns

	int j = OsMod(OsRand(), gcNonEmptyColumns);
	for (int i = 0; i < kcInvaderColumns; i++) {
		if (ganRowLastInvader[i] == -1)
			continue;
		if (j == 0)
			break;
		j--;
	}

	// i is the invader column we're interested in. Figure out
	// the x, y of the bottom invader in this column

	int x;
	int y;
	CalcInvaderPos(i, ganRowLastInvader[i], &x, &y);

	// Set up this missle for firing. Keep one y invader move
	// between invader and missle so if invader moves down a row
	// it doesn't run into missle.

	Missle *pmsl = &gamsl[nMslReady];
	memset(pmsl, 0, sizeof(Missle));
	pmsl->nState = kMslAnimating;
	pmsl->x = x + (kcxInvader - kcxInvaderMissle) / 2;
	pmsl->y = y + kcyInvader + kcyInvaderMove;
	pmsl->dy = kdyInvaderMissleMove;
	pmsl->msMove = kmsInvaderMissleMove;
	pmsl->msMoveAcc = kmsInvaderMissleMove;
	pmsl->cBitmaps = 3;
	pmsl->cBitmapsTrail = 4;
	pmsl->bid = kbidMissleFrame1;

	// Quick check: the missle can start below the ship when the invader is
	// on the last row right above the ship. Check for this case and start
	// the ship exploding if true.

	if (pmsl->y >= kyShip + kcyShip && pmsl->x + kcxInvaderMissle >= gxShip &&
			pmsl->x < gxShip + kcxShip) {
		// Explode ship, turn missle off

		ExplodeMotherShip(TRUE);
		pmsl->nState = kMslNone;
	}
}

void StartMotherShip()
{
	// If the mothership isn't moving, wait for the right time

	if (gxMotherShip != -1)
		return;

	gmsMotherShipWaitAcc += kmsOnTimer;
	if (gmsMotherShipWaitAcc < kmsMotherShipWait)
		return;

	// If top row of invaders overlaps with mothership, then don't
	// do mothership (diverges from actual game, but necessary)

	if (gyTopInvader < kyMotherShip + kcyMotherShip) {
		gmsMotherShipWaitAcc = kmsMotherShipWait;
		return;
	}

	// Reset wait time

	gmsMotherShipWaitAcc = 0;

	// forward or backwards?
	
	if (OsRand() & 1) {
		gxMotherShip = kxScreenLeft;
		gdxMotherShipMove = kdxMotherShipMove;
	} else {
		gxMotherShip = kxScreenRight - kcxMotherShip;
		gdxMotherShipMove = -kdxMotherShipMove;
	}

	// Clear out the lane because the score may be here

	OsFill(kxScreenLeft, kyMotherShip, kcxScreen, kcyMotherShip, FALSE);

	// Turn on the mothership walk sound

	OsPlaySound(kSndMotherWalk);
}


void MoveMotherShip()
{
	// If mothership hasn't started, just return

	if (gxMotherShip == -1)
		return;

	// Time movement

	gmsMotherShipMoveAcc += kmsOnTimer;
	if (gmsMotherShipMoveAcc < kmsMotherShipMove)
		return;
	gmsMotherShipMoveAcc -= kmsMotherShipMove;
	gxMotherShip += gdxMotherShipMove;

	// If still on screen, then draw

	if (gxMotherShip >= kxScreenLeft && gxMotherShip <= kxScreenRight - kcxMotherShip) {
		OsBlt(kbidMotherShip, gxMotherShip, kyMotherShip);
		OsBlt(kbidMotherShipStrip1 + OsMod(gxMotherShip, kcMotherShipAnims),
				gxMotherShip + kxMotherShipAnim,
				kyMotherShipAnim + kyMotherShip);
		return;
	}

	// Turn mothership off

	gxMotherShip = -1;
	OsFill(kxScreenLeft, kyMotherShip, kcxScreen, kcyMotherShip, FALSE);
	OsPlaySound(kSndMotherOff);
}

void DrawScore(bool fOverride)
{
	if (!fOverride) {
		// Nothing to do if don't need to update score

		if (!gfUpdateScore)
			return;

#if 0
		// Can't update if invaders are here

		if (gyTopInvader < kyScore + kcyScore)
			return;
#endif
	}

	// Zero fill only because of variable pitch fonts
	
	gfUpdateScore = FALSE;

	OsDrawScore();
}

void GamePlay()
{
	// Don't do these things if ship is exploding

	if (gcShipExplosionFrames == 0) {
		// Only if ship isn't exploding: if out of invaders,
		// advance to next level

		if (gnBottomRow < gnTopRow) {
			gnState = knLevelEnd;
			return;
		}

		MoveShip();
		MoveInvaders();
		PlayWalkSound();
		FireShipMissle();
		FireInvaderMissles();
		StartMotherShip();
	}

	// Do these things always

	MoveMotherShip();
	MoveMissles();
	ExplodeShip(FALSE);
	ExplodeMotherShip(FALSE);
	DrawScore(FALSE);
}

void GameEnd()
{
	OsDrawGameOver();

	// Wait 4 seconds before going into demo mode

	OsSleep(4000);

	// Game end sound

	if (gnState != knDemoMode)
		OsPlaySound(kSndGameEnd);

	// Go into demo mode

	gnDemoState = knDemoInit;
	gnState = knDemoMode;
}

void LevelEnd()
{
	// Do something here?
	
	gnState = knLevelStart;
}

void DemoInit()
{

	OsDrawDemoInit();

	// Put up score!

	gunScore = 0;
	InsertDecimalAscii(&gszScore[gnIndexScore], gunScore, 5);
	InsertDecimalAscii(&gszHighScore[gnIndexHighScore], gunHighScore, 5);
	DrawScore(TRUE);

	// Go to next mode

	gmsDemoHelpAcc = 0;
	gnDemoState = knDemoHelp;
}

void DemoHelp()
{
	gmsDemoHelpAcc += kmsOnTimer;
	if (gmsDemoHelpAcc > kmsDemoHelp) {
		gmsDemoHelpAcc = 0;

		// Start the invader demo

		GameStart();
		LevelStart();
		gcShips = 1;
		gnState = knDemoMode;
		gnDemoState = knDemoInvaders;
		return;
	}
}

void DemoInvaders()
{
	static int cShipShots;
	static int cShipMoves;
	static int fusInput;

	// Figure out input the ship should receive

	if ((cShipShots <= 0 && cShipMoves <= 0) || gxShip <= kxLeftShipBoundary ||
			gxShip >= kxRightShipBoundary) {
		// Figure out if we want to shoot (1 out of 3)

		fusInput = 0;
		if (OsMod(OsRand(), 3) == 0) {
			// Shoot at least once, up to 3 times

			cShipShots = OsMod(OsRand(), 3) + 1;
			fusInput |= kfFireKey;
		} else {
			cShipMoves = OsMod(OsRand(), 37);
			if (OsRand() & 1) {
				fusInput |= kfLeftKey;
			} else {
				fusInput |= kfRightKey;
			}
		}
	}

	// Old ship position

	int xShip = gxShip;

	// Play the game with this input

	gfusInput = fusInput;
	GamePlay();

	// Adjust our counters

	cShipMoves--;
	if (cShipShots != 0) {
		if (!(gfusInput & kfFireKey)) {
			cShipShots--;
			fusInput |= kfFireKey;
		}
	}
	gfusInput = 0;

	// If the ship was killed, go back to the help screen
	
	if (gnState == knGameEnd) {
		gnState = knDemoMode;
		gnDemoState = knDemoInit;
	}
}

void DemoMode()
{
	// If any keys being pressed, start game

	if (gfusInput & (kfLeftKey | kfRightKey | kfFireKey)) {
		gnState = knGameStart;
		return;
	}

	// Either showing the instruction screen or moving invaders

	switch (gnDemoState) {
	case knDemoInit:
		DemoInit();
		break;

	case knDemoHelp:
		DemoHelp();
		break;

	case knDemoInvaders:
		DemoInvaders();
		break;
	}
}

void OnTimer()
{
	// Use switch statement for now since the pilot can't handle
	// fixups yet

	switch (gnState) {
	case knDemoMode:
		DemoMode();
		break;

	case knGameStart:
		GameStart();
		break;

	case knLevelStart:
		LevelStart();
		break;

	case knGamePlay:
		GamePlay();
		break;

	case knGameEnd:
		GameEnd();
		break;

	case knLevelEnd:
		LevelEnd();
		break;

	default:
		break;
	}
}
