/////////////////////////////////////////////////////////////////////////////
// Invaders.h
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(__INVADERS_H__)
#define __INVADERS_H__

#include <tchar.h>

typedef unsigned long ulong;
typedef unsigned char byte;
typedef unsigned short ushort;

// Apple Invader constants (very close to original Space Invaders)
// Gleaned by visual inspection...
//
// - 200 pixels across!
// - Invaders 11 x 5 rows / columns
// - Invaders 16 x 8 pixels
// - Invaders move left-right by 2 pixels. 2 pixels of black on each invader
//   for auto-erasing (so it appears as if the invaders are 4 pixels apart)
// - Invaders move top-bottom by 8 pixels.
// - Invader rows separated by 8 pixels.
// - 11 columns make 14 moves left or right before moving down one row.
// - Invaders start 9 moves in, so 5 initial right moves before moving down.
// - Shields 21 x 16.
// - Ship 15 x 8. x pixel on when ship left most: 9. Same on other side.
//   pixels on each side for auto-erasing.
// - Platform underneath ship, 200 x 2 pixels.
// - Six pixels horz between platform and ship, 8 between ship and base, 48
//   between base and bottom row (level 1 initial position of invaders).
// - 22 pixels between left screen edge and left shield, 24 pixels between
//   shields (4 shields total) followed by 22 pixels between last shield and
//   right edge of screen (200 pixels total).
// - 2.5 seconds to move ship from far left to far right
// - 14 seconds for level 1 invader column to move from far left to far right
//   (doesn't include the downward movement)
// - 5 seconds for mother ship to move across screen

int const kyScreenTop = 0;
int const kyScreenBottom = 234;
int const kxScreenLeft = 0;
int const kxScreenRight = 400;
int const kcxScreen = 400;
int const kcyScreen = 234;

int const kcxBorder = 5;
int const kcyBorder = 3;
int const kxInvaderWindow = kcxBorder;
int const kyInvaderWindow = kcyBorder;
int const kxMarginLeft = kxInvaderWindow + kcxScreen + kcxBorder;
int const kyMarginTop = kyInvaderWindow;
int const kyMarginBottom = 240 - kcyBorder;
int const kcxMargin = 480 - kcxBorder - kxMarginLeft;
int const kxMarginRight = kxMarginLeft + kcxMargin;

// # of invader columns

int const kcInvaderColumns = 11;

// # of invader rows (don't change this)

int const kcInvaderRows = 5;

// cx & cy of invaders

int const kcxInvader = 32;
int const kcyInvader = 16;

// Initial x of top left invader (at the start of all levels)

int const kxInvaderStart = 32 + kxScreenLeft;

// Initial y of top left invader, first level

int const kyInvaderLevel1 = 0 + kyScreenTop;

// Maximum y starting position of top level invader, any level

int const kyInvaderStartMax = 96 + kyScreenTop;

// Horz pixels between invaders

int const kcxInvaderInvaderSpace = 0;

// Vert pixels between invaders

int const kcyInvaderInvaderSpace = 8;

// Horz pixels per horz invader move
// Vert pixels per vert invader move

int const kcxInvaderMove = 4;
int const kcyInvaderMove = 10;

// Invader playfield boundaries

int const kxLeftInvaderBoundary = -4 + kxScreenLeft;
int const kxRightInvaderBoundary = 404 + kxScreenLeft;
int const kyTopInvaderBoundary = 0 + kyScreenTop;
int const kyBottomInvaderBoundary = 192 + kyScreenTop;

// x, y, cx, cy of platform

int const kcyPlatform = 4;
int const kcxPlatform = 400;
int const kxPlatform = 0 + kxScreenLeft;
int const kyPlatform = kcyScreen - kcyPlatform + kyScreenTop;

// Pixels between platform and bottom of ship

int const kcyShipPlatformSpace = 12;

// Pixels between top of ship and bottom of shield

int const kcyShieldShipSpace = 16;

// Initial pixels between top of shield and bottom of last invader row

int const kcyShieldInvaderSpace = 48;

// # of shields

int const kcShields = 4;

// x & y of first shield

int const kxShieldFirst = 44 + kxScreenLeft;
int const kyShieldFirst = 154 + kyScreenTop;

// Horz space between shields

int const kcxShieldShieldSpace = 48;

// Time between ticks

int const kmsOnTimer = 30;

// Time it takes between invader row moves. Increases with
// every invader kill. Important stats:
// - 1000 ms to move 55 invaders one step
// - 280 ms to move 7 invaders one step (so that missles from top invaders
//   don't hit bottom invaders)
// - 11 ms to move 1 invader one step (so that 1 invader takes 1 second to
//   span screen on 200 pixel device)
// Fit the equation y = -0.5525x2 + 49.253x - 37.701, where x == # of
// invaders. Below grossly fits this in a simple way, but looks natural
// during game play.

int const kmsMoveTotal = 1005;
int const kmsSpeedupInit = 22;
int const kmsSpeedupInc = 1;
int const kcInvadersSpeedupAfter = 29;

// Time it takes to between invader walk sounds
// Start 750 apart

int const kmsWalkSoundInit = 750;
int const kmsWalkSoundSpeedup = 11;

// Time ship takes to move one pixel

int const kmsShipMove = 24;
int const kdxShipMove = 4;

// y, cx of ship

int const kyShip = 202 + kyScreenTop;
int const kcxShip = 34;
int const kcyShip = 16;

// cx, cy of ship explosion

int const kcxShipExplosion = 34;
int const kcyShipExplosion = 22;

// Amount ship missle moves per move

int const kdyShipMissleMove = 4;

// Time between ship missle moves

int const kmsShipMissleMove = 7;

// Amount to adjust explosion caused by ship missle

int const kdxShipMissleExplosionAdjust = 4;

// Dimensions of ship missle explosion (when not an invader)

int const kcxShipMissleExplosion = 16;
int const kcyShipMissleExplosion = 16;

// Time that an explosion stays on the screen

int const kmsExplosionTimeout = 250;

// Dimensions of ship missle

int const kcxShipMissle = 2;
int const kcyShipMissle = 4;

// # of simultaneous invader missles

int const kcInvaderMissles = 3;

// # pixels to move invader missle

int const kdyInvaderMissleMove = 4;

// Speed of invader missle

int const kmsInvaderMissleMove = 30;

// Width & height of invader missle explosion

int const kcxInvaderMissleExplosion = 16;
int const kcyInvaderMissleExplosion = 14;

// Width, height of invader missle

int const kcxInvaderMissle = 6;
int const kcyInvaderMissle = 4;

// Invader fire every # of ticks

int const kcTicksPerInvaderFire = 5;

// Left, right most pixel that ship can be at

int const kxLeftShipBoundary = 14 + kxScreenLeft;
int const kxRightShipBoundary = 356 + kxScreenLeft;

// # of ships to start with

int const kcShipsStart = 3;

// Time between ship explosion frames

int const kmsShipExplosion = 200;

// # of times to animate ship explosion

int const kcShipExplosionFrames = 10;

// Mothership constants

int const kcxMotherShip = 40;
int const kcyMotherShip = 16;
int const kmsMotherShipMove = 70;
int const kdxMotherShipMove = 4;
int const kyMotherShip = 0 + kyScreenTop;
int const kmsMotherShipWait = 23000;
int const kyMotherShipAnim = 8;
int const kxMotherShipAnim = 6;
int const kcxMotherShipAnim = 24;
int const kcMotherShipAnims = 7;
int const kmsMotherShipExplosion = 2000;

// Where to put the score

int const kxScore = -100;
int const kyScore = 0 + kyScreenTop;
int const kcyScore = 14;
int const kcxScore = 1;
int const kcxOneDigit = 14;
int const kcyOneDigit = 14;
int const kxHighScore = -100;
int const kxShipsLeft = 411;
int const kyShipsLeft = 240 - kcyBorder - kcyShip - 2;
int const kxShipsLeftCount = -100;
int const kyShipsLeftCount = -100;

// Help screen in demo mode timeout

int const kmsDemoHelp = 6000;

// Help button positions

int const kcxBtn = 0;
int const kcyBtn = 0;
int const kyBtn = kyScreenBottom - kcyBtn;

// Missle structure.

int const kMslNone = 0;
int const kMslAnimating = 1;
int const kMslExploding = 2;

struct Missle // msl
{
	int x;
	int y;
	int dy;
	int msMove;
	int msMoveAcc;
	byte nState;
	byte nBitmap;
	byte cBitmaps;
	byte cBitmapsTrail;
	byte cBitmapsErase;
	byte bid;
};

typedef struct _ByteRectangle { // brc
	char x;
	char y;
	unsigned char cx;
	unsigned char cy;
} ByteRectangle;

// Bitmaps we have

#define kbidBulletExplosion             0
#define kbidInvaderExplosion            1
#define kbidBottomInvaderClosed         2
#define kbidBottomInvaderOpen           3
#define kbidMiddleInvaderClosed         4
#define kbidMiddleInvaderOpen           5
#define kbidTopInvaderClosed            6
#define kbidTopInvaderOpen              7
#define kbidMissleFrame1                8
#define kbidMissleFrame2                9
#define kbidMissleFrame3                10
#define kbidShipMissle                  11
#define kbidMissleExplosion             12
#define kbidMotherShip                  13
#define kbidMotherShipStrip1            14
#define kbidMotherShipStrip2            15
#define kbidMotherShipStrip3            16
#define kbidMotherShipStrip4            17
#define kbidMotherShipStrip5            18
#define kbidMotherShipStrip6            19
#define kbidMotherShipStrip7            20
#define kbidShipExplosionFrame1         21
#define kbidShipExplosionFrame2         22
#define kbidShield                      23
#define kbidShip                        24
#define kbidMiniShip					25
int const kcBitmaps = 26;

// Sounds

int const kSndInvaderWalk0 = 0;
int const kSndInvaderWalk1 = 1;
int const kSndInvaderWalk2 = 2;
int const kSndInvaderWalk3 = 3;
int const kSndFire = 4;
int const kSndInvaderHit = 5;
int const kSndMotherHit = 6;
int const kSndMotherWalk = 7;
int const kSndBaseHit = 8;
int const kSndMotherOff = 9;
int const kSndGameStart = 10;
int const kSndGameEnd = 11;
int const kSndPurge = 12;

// Game state

int const knDemoMode = 0;
int const knGameStart = 1;
int const knLevelStart = 2;
int const knGamePlay = 3;
int const knGameEnd = 4;
int const knLevelEnd = 5;

// Demo State

int const knDemoInit = 0;
int const knDemoHelp = 1;
int const knDemoInvaders = 2;

// Os calls

extern int gcShips;
extern unsigned int gunHighScore;
extern ByteRectangle gabrc[kcBitmaps];
extern TCHAR gszScore[];
extern TCHAR gszHighScore[];
extern int gnState;
extern int gnDemoState;

void OsFill(int x, int y, int cx, int cy, bool fWhite);
#define OsBlt(bid, xDst, yDst) \
		OsBltHelper(bid, xDst, yDst, 0)
#define OsOrBlt(bid, xDst, yDst) \
		OsBltHelper(bid, xDst, yDst, 4)
#define OsNotAndBlt(bid, xDst, yDst) \
		OsBltHelper(bid, xDst, yDst, 2)
void OsBltHelper(int bid, int xDst, int yDst, int rop);
bool OsCheckBits(int x, int y, int cx, int cy, bool *pfOffScreen);
int OsRand();
void OsSleep(int ms);
void OsTextOut(int x, int y, TCHAR *psz);
short OsDiv(long sDividend, short sDivisor);
short OsMod(long sDividend, short sDivisor);
int OsTextExtent(TCHAR *psz);
void OsDrawHelpButtons();
void OsPlaySound(int nSound);
void OsDrawScore();
void OsDrawGameOver();
void OsDrawDemoInit();

// Invader callback

ushort const kfLeftKey = 0x0008;
ushort const kfRightKey = 0x0010;
ushort const kfFireKey = 0x0040;
ushort const kfWaitFireKeyUp = 0x8000;
extern ushort gfusInput;

void OnTimer();

#endif // !defined(__INVADERS_H__)