/////////////////////////////////////////////////////////////////////////////
// Os.cpp
//
// All Os specific support goes in here, including main().
/////////////////////////////////////////////////////////////////////////////

#include <Windows.h>
#include "Invaders.h"
#include "Resource.h"

HWND ghwnd;
HWND ghwndMain;
HDC ghdc;
HDC ghdcMem;
HBITMAP ghbmCheckBits;
HBITMAP ghbmSav;
HBITMAP ghbmAll;
ushort gfusInput;
HINSTANCE ghInst;
void *gpvCheckBits;
HFONT gahfnt[3];
BOOL gfActive;
BOOL gfTopMost;
BOOL gfSoundEffects = TRUE;

void PaintParent(HDC hdc);
void CheckKeys(MSG *pmsg);
void CheckKeyClick();

// From C runtimes

int holdrand;
void __cdecl srand (unsigned int seed)
{
   holdrand = (long)seed;
}

int __cdecl rand (void)
{
	return(((holdrand = holdrand * 214013L + 2531011L) >> 16) & 0x7fff);
}

wchar_t * __cdecl iwcscpy(wchar_t * dst, const wchar_t * src)
{
        wchar_t * cp = dst;

        while( *cp++ = *src++ )
                ;               /* Copy src over dst */

        return( dst );
}

size_t __cdecl iwcslen (
        const wchar_t * wcs
        )
{
        const wchar_t *eos = wcs;

        while( *eos++ ) ;

        return( (size_t)(eos - wcs - 1) );
}

// This is the maximum area checked for missle collisions. Since on Win32
// bitmaps are ulong aligned, our bitmap bits are kcyCheckBits ulongs.

int const kcxCheckBits = 6;
int const kcyCheckBits = 6;

short OsDiv(long sDividend, short sDivisor)
{
	return sDividend / sDivisor;
}

short OsMod(long sDividend, short sDivisor)
{
	return sDividend % sDivisor;
}

void OsFill(int x, int y, int cx, int cy, bool fWhite)
{
	PatBlt(ghdc, x, y, cx, cy, !fWhite ? WHITENESS : BLACKNESS);
}

// Masks. Pegasus can't do NOT AND so we need to have
// separate mask images

char mpbidbidMask[] = {
	0,
	1,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1,
	2,
	3,
	4,
	5,
	6,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1,
	-1
};

// Invader bitmap positions & sizes

ByteRectangle gabrc[kcBitmaps] = {
	{ 74, 32, 16, 16 }, // BulletExplosion
	{ 64, 16, 30, 16 }, // InvaderExplosion
	{ 0, 0, 32, 16 }, // BottomInvaderClosed
	{ 32, 0, 32, 16 }, // BottomInvaderOpen
	{ 96, 0, 32, 16 }, // MiddleInvaderClosed
	{ 64, 0, 32, 16 }, // MiddleInvaderOpen
	{ 0, 16, 32, 16 }, // TopInvaderClosed
	{ 32, 16, 32, 16 }, // TopInvaderOpen
	{ 108, 32, 6, 4 }, // MissleFrame1
	{ 108, 36, 6, 4 }, // MissleFrame2
	{ 108, 40, 6, 4 }, // MissleFrame3
	{ 104, 56, 2, 3 }, // ShipMissle
	{ 90, 32, 16, 16 }, // MissleExplosion
	{ 34, 32, 40, 16 }, // MotherShip
	{ 0, 78, 24, 2 }, // MotherShipStrip1
	{ 2, 78, 24, 2 }, // MotherShipStrip2
	{ 4, 78, 24, 2 }, // MotherShipStrip3
	{ 6, 78, 24, 2 }, // MotherShipStrip4
	{ 8, 78, 24, 2 }, // MotherShipStrip5
	{ 10, 78, 24, 2 }, // MotherShipStrip6
	{ 12, 78, 24, 2 }, // MotherShipStrip7
	{ 0, 48, 34, 22 }, // ShipExplosionFrame1
	{ 32, 48, 34, 22 }, // ShipExplosionFrame2
	{ 86, 48, 42, 32 }, // Shield
	{ 0, 32, 34, 16 }, // Ship
	{ 38, 72, 15, 8 } // Mini-ship
};

// Mask positions & sizes

ByteRectangle gabrcMask[] = {
	{ 66, 48, 16, 16 }, // BulletExplosion
	{ 96, 16, 30, 16 }, // InvaderExplosion
	{ 116, 32, 6, 4 }, // MissleFrame1
	{ 116, 36, 6, 4 }, // MissleFrame2
	{ 116, 40, 6, 4 }, // MissleFrame3
	{ 124, 32, 2, 6 }, // ShipMissle
	{ 66, 64, 16, 16 }, // MissleExplosion
};

void OsBltHelper(int bid, int xDst, int yDst, int rop)
{
	ByteRectangle *pbrc = &gabrc[bid];
	HBITMAP hbmSav = (HBITMAP)SelectObject(ghdcMem, (HBITMAP)ghbmAll);
	int bidT;

	switch (rop) {
	case 0:
		// ROP == SRCCOPY (dst = src)

		BitBlt(ghdc, xDst, yDst, pbrc->cx, pbrc->cy,
				ghdcMem, pbrc->x, pbrc->y, SRCCOPY);
		break;

	case 2:
		// ROP = SRCPAINT (dst = dst OR src)

		bidT = mpbidbidMask[bid];
		pbrc = &gabrcMask[bidT];
		BitBlt(ghdc, xDst, yDst, pbrc->cx, pbrc->cy,
				ghdcMem, pbrc->x, pbrc->y, SRCPAINT);
		break;

	case 4:
		// ROP == SRCAND

		BitBlt(ghdc, xDst, yDst, pbrc->cx, pbrc->cy,
				ghdcMem, pbrc->x, pbrc->y, SRCAND);
		break;
	}
	SelectObject(ghdcMem, hbmSav);
}

// Check if any bits if the destination are on. If
// so, return TRUE. Out of screen bounds counts as TRUE.

bool OsCheckBits(int x, int y, int cx, int cy, bool *pfOffScreen)
{
	// If out of bounds, then assume bits are on

	*pfOffScreen = FALSE;
	if (y < 0 || y + cy >= kcyScreen) {
		*pfOffScreen = TRUE;
		return TRUE;
	}

	// Grab bits of destination

#define aulBits ((ulong *)gpvCheckBits)
//	memset(aulBits, 0xff, sizeof(ulong) * 6);

	HBITMAP hbmSav = (HBITMAP)SelectObject(ghdcMem, ghbmCheckBits);
	BitBlt(ghdcMem, 0, 0, cx, cy, ghdc, x, y, SRCCOPY);
	SelectObject(ghdcMem, hbmSav);

	ushort us = 0;
	switch (cx) {
	case 2:
		us = 0x00f0;
		break;

	case 6:
		us = 0xf0ff;
		break;
	}

	for (int i = 0; i < cy; i++) {
		if ((LOWORD(aulBits[i]) & us) != us)
			return TRUE;
	}

	return FALSE;
}

void OsSetFont(int n)
{
	SelectObject(ghdc, gahfnt[n]);
}

void OsTextOut(int x, int y, TCHAR *psz)
{
	ExtTextOut(ghdc, x, y, 0, NULL, psz, iwcslen(psz), NULL);
}

int OsTextExtent(TCHAR *psz)
{
	SIZE siz;
	GetTextExtentExPoint(ghdc, psz, iwcslen(psz), 1000, NULL, NULL, &siz);
	return (int)siz.cx;
}

void *gpvWav[9];

void InitSound()
{
	for (int i = 0; i < 9; i++) {
		HRSRC hr = FindResource(ghInst, MAKEINTRESOURCE(i), MAKEINTRESOURCE(55));
		HGLOBAL hg = LoadResource(ghInst, hr);
		gpvWav[i] = LockResource(hg);
	}
}

void OsPlaySound(int nSound)
{
	extern int gxMotherShip;

	if (gnState == knDemoMode)
		return;

	if (!gfSoundEffects)
		return;

	if (!gfActive)
		return;

	switch (nSound) {
	case kSndGameStart:
	case kSndGameEnd:
		break;

	case kSndMotherWalk:
		PlaySound((const unsigned short *)gpvWav[nSound], NULL, SND_MEMORY | SND_ASYNC | SND_LOOP);
		break;

	case kSndMotherOff:
		PlaySound(NULL, NULL, 0);
		break;

	case kSndInvaderWalk0:
	case kSndInvaderWalk1:
	case kSndInvaderWalk2:
	case kSndInvaderWalk3:
		// SND_NOWAIT makes these sounds lower priority than any playing sound

//		PlaySound((const unsigned short *)gpvWav[nSound], NULL, SND_MEMORY | SND_ASYNC | SND_NOWAIT | SND_NOSTOP);
		break;

	case kSndFire:
	case kSndInvaderHit:
		// Don't play these if the mothership is walking

		if (gxMotherShip == -1)
			PlaySound((const unsigned short *)gpvWav[nSound], NULL, SND_MEMORY | SND_ASYNC);
		break;

	default:
		PlaySound((const unsigned short *)gpvWav[nSound], NULL, SND_MEMORY | SND_ASYNC);
		break;
	}
}

void OsDrawHelpButtons()
{
#if 0
	HBITMAP hbm = LoadBitmap(ghInst, MAKEINTRESOURCE(kbidButtons));
	HBITMAP hbmSav = SelectObject(ghdcMem, (HBITMAP)hbm);

	BitBlt(ghdc, kxBtn1, kyBtn, kcxBtn, kcyBtn,
			ghdcMem, 0, 0, NOTSRCCOPY);
	BitBlt(ghdc, kxBtn2, kyBtn, kcxBtn, kcyBtn,
			ghdcMem, kcxBtn, 0, NOTSRCCOPY);
	BitBlt(ghdc, kxBtn3, kyBtn, kcxBtn, kcyBtn,
			ghdcMem, kcxBtn * 2, 0, NOTSRCCOPY);

	SelectObject(ghdcMem, (HBITMAP)hbmSav);
#endif
}

#define kcxShipShipSpace (-4)
#define kcyShipScoreSpace 4

int CalcScoreCy(SIZE *psiz1, SIZE *psiz2)
{
	// Calc score sizes

	int cy = 0;

	HFONT hfntT = (HFONT)SelectObject(ghdc, gahfnt[0]);
	GetTextExtentExPoint(ghdc, TEXT("H"), 1, 1000, NULL, NULL, psiz1);
	cy += psiz1->cy * 2;

	SelectObject(ghdc, gahfnt[2]);
	GetTextExtentExPoint(ghdc, TEXT("00000"), 5, 1000, NULL, NULL, psiz2);
	cy += psiz2->cy * 2;

	cy += kcyShipScoreSpace + kcyShip + 2;
	SelectObject(ghdc, hfntT);

	return cy;
}

void OsDrawScore()
{
	// Drawing in the parent window now

	HDC hdc = GetDC(ghwndMain);
	HDC hdcT = ghdc;
	ghdc = hdc;

	// Calc score sizes

	SIZE siz1;
	SIZE siz2;

	int y = kcyBorder;

	// Draw scores 

	CalcScoreCy(&siz1, &siz2);

	SelectObject(hdc, gahfnt[0]);
	ExtTextOut(hdc, kxMarginLeft + 2, y, 0, NULL, TEXT("Score"), 5, NULL);
	y += siz1.cy;
	SelectObject(hdc, gahfnt[2]);
	ExtTextOut(hdc, kxMarginLeft + (kcxMargin - siz2.cx) / 2, y, 0, NULL, gszScore, iwcslen(gszScore), NULL);
	y += siz2.cy;
	SelectObject(hdc, gahfnt[0]);
	ExtTextOut(hdc, kxMarginLeft + 2, y, 0, NULL, TEXT("High Score"), 10, NULL);
	y += siz1.cy;
	SelectObject(hdc, gahfnt[2]);
	ExtTextOut(hdc, kxMarginLeft + (kcxMargin - siz2.cx) / 2, y, 0, NULL, gszHighScore, iwcslen(gszHighScore), NULL);
	y += siz2.cy;

	y += kcyShipScoreSpace;

	HBRUSH hbr = CreateSolidBrush(0x00FFFFFF);
	RECT rc;
	rc.left = kxMarginLeft;
	rc.top = y;
	rc.right = kxMarginLeft + kcxMargin;
	rc.bottom = y + kcyShip;
	FillRect(hdc, &rc, hbr);
	DeleteObject(hbr);

	// Draw # of ships left

	int x = rc.left;
	for (int i = 1; i < gcShips; i++) {
		OsOrBlt(kbidShip, x, y);
		x += kcxShip + kcxShipShipSpace;
	}

	ghdc = hdcT;
	ReleaseDC(ghwndMain, hdc);
}

TCHAR gszGameOver[] = TEXT("Game Over");

void OsDrawGameOver()
{
	// Figure out the length of Game Over

	int cx = OsTextExtent(gszGameOver);

	// Clear rect in middle of screen, with a one char clear boundary
	// around

	int x = OsDiv((kcxScreen - (cx + kcyOneDigit * 2)), 2);
	int y = (kyScreenBottom - kcyOneDigit * 3) / 2;
	OsFill(x, y, cx + kcyOneDigit * 2, kcyOneDigit * 3, FALSE);

	// Output Game Over

	OsTextOut(x + kcyOneDigit, y + kcyOneDigit, gszGameOver);

	RECT rc;
	SetRect(&rc, x, y, x + cx + kcyOneDigit * 2, y + kcyOneDigit * 3);
	InflateRect(&rc, 2, 2);
	DrawEdge(ghdc, &rc, EDGE_BUMP, BF_RECT);
}

TCHAR gszMysteryPoints[] = TEXT(" ? Mystery");
TCHAR gsz30Points[] = TEXT(" 30 Points");
TCHAR gsz20Points[] = TEXT(" 20 Points");
TCHAR gsz10Points[] = TEXT(" 10 Points");
TCHAR gszAnyKey[] = TEXT("Press Space Bar To Start");
TCHAR gszSpaceInvaders[] = TEXT("Invaders");
TCHAR gszScottLudwig[] = TEXT("by Scott Ludwig (http://www.eskimo.com/~scottlu)");

void OsDrawDemoInit()
{
	// Erase screen

	OsFill(kxScreenLeft, kyScreenTop, kcxScreen, kcyScreen, FALSE);

	// Put up "Space Invaders"

	LOGFONT lf;
	memset(&lf, 0, sizeof(lf));
	lf.lfWeight = FW_BOLD;
	lf.lfCharSet = ANSI_CHARSET;
	lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
	lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	lf.lfQuality = DEFAULT_QUALITY;
	lf.lfPitchAndFamily = DEFAULT_PITCH | FF_MODERN;
	iwcscpy(lf.lfFaceName, TEXT("Times New Roman"));

	lf.lfHeight = 32;
	HFONT hfnt = CreateFontIndirect(&lf);

	if (hfnt == NULL)
		OsSetFont(2);
	HFONT hfntT = (HFONT)SelectObject(ghdc, hfnt);
	SIZE siz;
	GetTextExtentExPoint(ghdc, gszSpaceInvaders, iwcslen(gszSpaceInvaders),
			9999, NULL, NULL, &siz);
//	int y = kcyScreen / 2 - siz.cy / 2;
	int y = 10;
	OsTextOut(OsDiv(kcxScreen - siz.cx, 2) + kxScreenLeft, y, gszSpaceInvaders);
//	y = 10 + siz.cy + (90 - (10 + siz.cy)) / 2;
//	y = 50;
	y = 10 + siz.cy + kcyOneDigit;
	SelectObject(ghdc, hfntT);
	DeleteObject(hfnt);

	// Put up "by Scott Ludwig"
    
	OsSetFont(1);
	int cx = OsTextExtent(gszScottLudwig);
	OsTextOut(OsDiv(kcxScreen - cx, 2) + kxScreenLeft, y, gszScottLudwig);

	// Put up invaders

//	y = (kcyScreen - (y + kcyOneDigit)) / 2 - kcyInvader * 2 - kcyOneDigit;
	y = 90;
	int x = kxScreenLeft + kcxMotherShip + OsDiv(kcxScreen - (kcxMotherShip +
			OsTextExtent(gszMysteryPoints)), 2);

	OsSetFont(0);
	OsBlt(kbidMotherShip, x - kcxMotherShip, y);
	OsTextOut(x, y + 1, gszMysteryPoints);
	y += kcyMotherShip + kcyOneDigit / 2;

	OsBlt(kbidTopInvaderClosed, x - kcxInvader - (kcxMotherShip -
			kcxInvader) / 2, y);
	OsTextOut(x, y + 1, gsz30Points);
	y += kcyInvader + kcyOneDigit / 2;

	OsBlt(kbidMiddleInvaderClosed, x - kcxInvader - (kcxMotherShip -
			kcxInvader) / 2, y);
	OsTextOut(x, y + 1, gsz20Points);
	y += kcyInvader + kcyOneDigit / 2;

	OsBlt(kbidBottomInvaderOpen, x - kcxInvader - (kcxMotherShip -
			kcxInvader) / 2, y);
	OsTextOut(x, y + 1, gsz10Points);

	// Put up "Press Any Key To Start"

	OsSetFont(1);
	y = kyScreenBottom - kcyOneDigit * 2;
	OsTextOut(OsDiv(kcxScreen - OsTextExtent(gszAnyKey), 2) + kxScreenLeft, y, gszAnyKey);

	// Draw help buttons

	OsDrawHelpButtons();

	OsSetFont(1);
}

int OsRand()
{
	return (int)rand();
}

void OsSleep(int ms)
{
	Sleep(ms);
}

#define Assert(f, sz) if (f == 0) MessageBox(NULL, TEXT(sz), TEXT("Error"), MB_OK);

typedef struct _BmpHeader {
	BITMAPINFOHEADER bih;
	RGBQUAD argb[4];
} BmpHeader;

bool Init()
{
	// Remember if the taskbar is topmost. We want to be on top of it!

	HWND hwndT = FindWindow(TEXT("HHTaskBar"), NULL);
	if (hwndT != NULL) {
		if (GetWindowLong(hwndT, GWL_EXSTYLE) & WS_EX_TOPMOST) {
			SetWindowPos(hwndT, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
			gfTopMost = TRUE;
		}
	}

	ghdc = GetDC(ghwnd);
	ghdcMem = CreateCompatibleDC(ghdc);

	// Create font
	
	LOGFONT lf;
	memset(&lf, 0, sizeof(lf));
	lf.lfWeight = FW_BOLD;
	lf.lfCharSet = ANSI_CHARSET;
	lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
	lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	lf.lfQuality = DEFAULT_QUALITY;
	lf.lfPitchAndFamily = DEFAULT_PITCH | FF_MODERN;
	iwcscpy(lf.lfFaceName, TEXT("Times New Roman"));

	lf.lfHeight = 12;
	gahfnt[0] = CreateFontIndirect(&lf);

	lf.lfHeight = 16;
	gahfnt[1] = CreateFontIndirect(&lf);

	lf.lfHeight = 24;
	gahfnt[2] = CreateFontIndirect(&lf);

	SelectObject(ghdc, gahfnt[1]);

	SetBkColor(ghdc, 0x00FFFFFF);
	SetTextColor(ghdc, 0x00000000);

	SetTextColor(ghdcMem, 0x00FFFFFF);
	SetBkColor(ghdcMem, 0x00000000);
	
	SetBkMode(ghdc, TRANSPARENT);

	BmpHeader bh;
	memset(&bh, 0, sizeof(bh));
	bh.bih.biSize = sizeof(bh.bih);
	bh.bih.biWidth = kcxCheckBits;
	bh.bih.biHeight = -kcyCheckBits;
	bh.bih.biPlanes = 1;
	bh.bih.biBitCount = 2;
	bh.bih.biCompression = BI_RGB;
	bh.argb[0].rgbBlue = bh.argb[0].rgbGreen = bh.argb[0].rgbRed = 0x00;
	bh.argb[1].rgbBlue = bh.argb[1].rgbGreen = bh.argb[1].rgbRed = 0x55;
	bh.argb[2].rgbBlue = bh.argb[2].rgbGreen = bh.argb[2].rgbRed = 0xaa;
	bh.argb[3].rgbBlue = bh.argb[3].rgbGreen = bh.argb[3].rgbRed = 0xFF;

	ghbmCheckBits = CreateDIBSection(ghdcMem, (BITMAPINFO *)&bh, DIB_RGB_COLORS, &gpvCheckBits, NULL, 0);

	
	if (ghdc == NULL || ghdcMem == NULL || ghbmCheckBits == NULL) {
		MessageBox(NULL, TEXT("Init() failed!"), TEXT("Error"), MB_OK);
		return FALSE;
	}

	// Create the master bitmap, blt into a like format
	// dib

#define kcxAll (128)
#define kcyAll (80)
	HBITMAP hbmT = LoadBitmap(ghInst, MAKEINTRESOURCE(IDB_BITMAP1));
	HDC hdcT = CreateCompatibleDC(ghdc);

	bh.bih.biWidth = kcxAll;
	bh.bih.biHeight = -kcyAll;

	void *pv;
	ghbmAll = CreateDIBSection(ghdcMem, (BITMAPINFO *)&bh, DIB_RGB_COLORS, &pv, NULL, 0);

	HBITMAP hbmT1 = (HBITMAP)SelectObject(hdcT, hbmT);
	HBITMAP hbmT2 = (HBITMAP)SelectObject(ghdcMem, ghbmAll);
	
	BitBlt(ghdcMem, 0, 0, kcxAll, kcyAll, hdcT, 0, 0, SRCCOPY);
	
	hbmT1 = (HBITMAP)SelectObject(hdcT, hbmT1);
	hbmT2 = (HBITMAP)SelectObject(ghdcMem, hbmT2);

	DeleteObject(hbmT);
	DeleteDC(hdcT);

	// Seed the random number generator

	srand(GetTickCount());

	// Grab the high score from the registry

	LRESULT lr;
	DWORD dwDisposition;
	HKEY hkey;
	DWORD cb;

	lr = RegCreateKeyEx(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Invaders"), 0, TEXT("High Score"),
			REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, &dwDisposition);
	if (lr == ERROR_SUCCESS && dwDisposition == REG_OPENED_EXISTING_KEY) {
		cb = sizeof(gunHighScore);
		lr = RegQueryValueEx(hkey, TEXT("High Score"), 0, NULL, (BYTE *)&gunHighScore, &cb);
		cb = sizeof(gfSoundEffects);
		lr = RegQueryValueEx(hkey, TEXT("Sound Effects"), 0, NULL, (BYTE *)&gfSoundEffects, &cb);
	}
	RegCloseKey(hkey);

	// Init the sound

	InitSound();

	return TRUE;
}

void Exit()
{	
	// Make the taskbar topmost if necessary

	HWND hwndT = FindWindow(TEXT("HHTaskBar"), NULL);
	if (hwndT != NULL && gfTopMost)
		SetWindowPos(hwndT, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);

	DeleteObject(ghbmCheckBits);
	DeleteObject(ghbmAll);
	DeleteDC(ghdcMem);
	DeleteObject(gahfnt[0]);
	DeleteObject(gahfnt[1]);
	DeleteObject(gahfnt[2]);

	// Save high score to the registry

	LRESULT lr;
	HKEY hkey;
	DWORD cb;

	lr = RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("Software\\Microsoft\\Invaders"), 0, KEY_ALL_ACCESS, &hkey);
	if (lr == ERROR_SUCCESS) {
		cb = sizeof(gunHighScore);
		lr = RegSetValueEx(hkey, TEXT("High Score"), 0, REG_NONE, (BYTE *)&gunHighScore, sizeof(gunHighScore));
		cb = sizeof(gfSoundEffects);
		lr = RegSetValueEx(hkey, TEXT("Sound Effects"), 0, REG_NONE, (BYTE *)&gfSoundEffects, sizeof(gfSoundEffects));
	}
	RegCloseKey(hkey);

}

void CheckKeyClick()
{
	LRESULT lr;
	HKEY hkey;
	DWORD cb;
	DWORD dwDisposition;

	lr = RegCreateKeyEx(HKEY_CURRENT_USER, TEXT("ControlPanel\\Volume"), 0, TEXT("Key"),
			REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, &dwDisposition);
	DWORD dwKeyClick = 0;
	cb = sizeof(dwKeyClick);
	lr = RegQueryValueEx(hkey, TEXT("Key"), 0, NULL, (BYTE *)&dwKeyClick, &cb);
	RegCloseKey(hkey);
	if (LOWORD(dwKeyClick) != 0) {
		MessageBox(ghwnd, TEXT("Invaders has determined that you have key click sounds turned on. Invaders \
sounds better if clicks are off. Start Menu->Settings->Control Panel->Volume & Sounds"), TEXT("Note"), MB_OK);

	}
}

#define kmsOnPegTimer (kmsOnTimer + 0)
int WINAPI WinMain(HINSTANCE hinst, HINSTANCE hinstPrev,
		TCHAR *pszCmdLine, int nCmdShow)
{
	extern LRESULT CALLBACK MainWndProc(HWND hwnd, UINT wm, WPARAM wp, LPARAM lp);
	extern LRESULT CALLBACK ParentWndProc(HWND hwnd, UINT wm, WPARAM wp, LPARAM lp);

	WNDCLASS wc;

	memset(&wc, 0, sizeof(wc));
	wc.lpfnWndProc = MainWndProc;
	wc.hInstance = hinst;
	wc.hCursor = NULL;
	wc.lpszClassName = TEXT("InvaderClass");
	wc.style = 0;
	wc.lpszMenuName = NULL;

	if (!RegisterClass(&wc)) {
Fail:
		MessageBox(NULL, TEXT("Initialization failure!"), TEXT("Error"), MB_OK);
		return 0;
	}

	wc.lpszClassName = TEXT("ParentInvaderClass");
	wc.lpfnWndProc = ParentWndProc;

	if (!RegisterClass(&wc))
		goto Fail;

	ghInst = hinst;

	int cxScreen = GetSystemMetrics(SM_CXSCREEN);
	int cyScreen = GetSystemMetrics(SM_CYSCREEN);

	// Create a window the size of the screen

	ghwndMain = CreateWindowEx(0, TEXT("ParentInvaderClass"), TEXT("Invaders"),
			WS_VISIBLE | WS_CLIPCHILDREN, 0, 0, cxScreen, cyScreen,
			NULL, NULL, hinst, NULL);
	if (ghwndMain == NULL)
		goto Fail;

	// Minimally this should be a 480x240 screen

	ghwnd = CreateWindowEx(0, TEXT("InvaderClass"), TEXT("Invaders"),
			WS_VISIBLE | WS_CHILD,	kxInvaderWindow, kyInvaderWindow, kcxScreen,
			kcyScreen, ghwndMain, NULL, hinst, NULL);
	if (ghwnd == NULL)
		goto Fail;

	// Create the exit button

	CreateWindowEx(0, TEXT("button"), TEXT("Exit"), WS_VISIBLE | WS_CHILD,
			kxMarginLeft, kyMarginBottom - kcyOneDigit * 2 - kcyBorder, kcxMargin, kcyOneDigit * 2,
			ghwndMain, (HMENU)1, hinst, NULL);

	// Create the sound effects on/off checkbox

	CreateWindowEx(0, TEXT("button"), TEXT("Sound"), WS_VISIBLE | WS_CHILD | BS_AUTOCHECKBOX,
			kxMarginLeft, kyMarginBottom - kcyOneDigit * 4 - kcyBorder, kcxMargin, kcyOneDigit * 2,
			ghwndMain, (HMENU)2, hinst, NULL);

	if (!Init())
		goto Fail;

	// Set state of checkbox now that we have restored the sound effects setting

	if (gfSoundEffects)
		SendDlgItemMessage(ghwndMain, 2, BM_SETCHECK, BST_CHECKED, 0);

	// Notify user if key clicks are turned on

	CheckKeyClick();

	MSG msg;
	long msSlop = 0;
	long msDelta = 0;
	while (TRUE) {
		// Start time since last tick, & update window

		long ms = GetTickCount();

		// Overcome weird bug on peg user. PeekMessage with no filter
		// will not process paint messages

		RECT rc;
		if (GetUpdateRect(ghwnd, &rc, FALSE) || GetUpdateRect(ghwndMain, &rc, FALSE))
			UpdateWindow(ghwndMain);

		// Process all messages

		while (TRUE) {
			if (gfActive) {
				if (!PeekMessage(&msg, NULL, 0, 0, TRUE))
					break;
			} else {
				GetMessage(&msg, NULL, 0, 0);
				
				// If became active, reset the clock

				if (gfActive)
					ms = GetTickCount();
			}

			if (msg.message == WM_QUIT) {
				PlaySound(NULL, NULL, 0);
				Exit();
				return 1;
			}
			CheckKeys(&msg);
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		// OnTimer() every kmsOnPegTimer milliseconds
		// LATER: deal with timer reset during modal operations

		long msT;
		while (TRUE) {
			msT = GetTickCount();
			msDelta = (msT - ms) + msSlop;
			msSlop = 0;

			if (msDelta >= kmsOnPegTimer)
				break;
//			Sleep(kmsOnPegTimer - msDelta);
		}

		while (msDelta >= kmsOnPegTimer) {
			msDelta -= kmsOnPegTimer;
			OnTimer();
		}

		// Remember how long it took to call OnTimer

		msSlop += (GetTickCount() - msT);
	}
}

LRESULT CALLBACK MainWndProc(HWND hwnd, UINT wm, WPARAM wp, LPARAM lp)
{
	switch (wm) {
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hwnd, &ps);
			if (!gfActive)
				OsDrawDemoInit();
			EndPaint(hwnd, &ps);
		}
		break;

	default:
		return DefWindowProc(hwnd, wm, wp, lp);
	}

	return 0;
}

void PaintParent(HDC hdc)
{
	HDC hdcT = hdc;
	if (hdc == NULL)
		hdcT = GetDC(ghwndMain);

	// Fill background with gray

	RECT rc;
	GetClientRect(ghwndMain, &rc);
	HBRUSH hbr = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
	FillRect(hdcT, &rc, hbr);
	DeleteObject(hbr);

	// Draw edge of invaders area

	GetWindowRect(ghwnd, &rc);
	MapWindowPoints(NULL, ghwndMain, (LPPOINT)&rc, 2);
	InflateRect(&rc, 2, 2);
	DrawEdge(hdcT, &rc, EDGE_BUMP, BF_RECT);

	// Draw score area

	SIZE siz1;
	SIZE siz2;
	rc.left = kxMarginLeft;
	rc.top = kyMarginTop;
	rc.right = kxMarginRight;
	rc.bottom = rc.top + CalcScoreCy(&siz1, &siz2);
	FillRect(hdcT, &rc, (HBRUSH)GetStockObject(WHITE_BRUSH));
	InflateRect(&rc, 2, 2);
	DrawEdge(hdcT, &rc, EDGE_BUMP, BF_RECT);

	if (hdc == NULL)
		ReleaseDC(ghwndMain, hdcT);

	OsDrawScore();
}

void CheckKeys(MSG *pmsg)
{
	switch (pmsg->message) {
	case WM_KEYDOWN:
		switch (pmsg->wParam) {
		case VK_LEFT:
			gfusInput |= kfLeftKey;
			break;

		case VK_RIGHT:
			gfusInput |= kfRightKey;
			break;

		default:
		case VK_SPACE:
//			if (!(gfusInput & kfWaitFireKeyUp))
				gfusInput |= kfFireKey;
			break;
		}
		break;

	case WM_KEYUP:
		switch (pmsg->wParam) {
		case VK_LEFT:
			gfusInput &= ~kfLeftKey;
			break;

		case VK_RIGHT:
			gfusInput &= ~kfRightKey;
			break;

		case VK_SPACE:
			gfusInput &= ~(kfFireKey | kfWaitFireKeyUp);
			break;
		}
		break;
	}
}

LRESULT CALLBACK ParentWndProc(HWND hwnd, UINT wm, WPARAM wp, LPARAM lp)
{
	HWND hwndT;

	switch (wm) {
	case WM_COMMAND:
		if (HIWORD(wp) == BN_CLICKED) {
			switch (LOWORD(wp)) {
			case 1:
				// user clicked on exit, so exit

				DestroyWindow(ghwndMain);
				break;

			case 2:
				// user turned sound on / off

				switch (SendDlgItemMessage(ghwndMain, 2, BM_GETSTATE, 0, 0) & (BST_CHECKED | BST_UNCHECKED)) {
				case BST_CHECKED:
					gfSoundEffects = TRUE;
					break;

				case BST_UNCHECKED:
					gfSoundEffects = FALSE;
					break;
				}

				if (!gfSoundEffects)
					PlaySound(NULL, NULL, 0);
				SetFocus(ghwnd);
				break;
			}
		}
		break;

	case WM_ACTIVATE:
		if (LOWORD(wp) == 0) {
			gfusInput = 0;

			// Go into demo mode

			gnDemoState = 0;
			gnState = 0;
			gcShips = 0;
			InvalidateRect(ghwnd, NULL, TRUE);
			gfActive = FALSE;

			// Put task bar on top - or topmost

			hwndT = FindWindow(TEXT("HHTaskBar"), NULL);
			if (hwndT != NULL)
				SetWindowPos(hwndT, HWND_TOP, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);

			// Turn sound off

			PlaySound(NULL, NULL, 0);
		} else {
			// Wake up out of GetMessage()
			// Redraw because not that we're active, the demo init
			// screen won't redraw on WM_PAINT demand

			gfActive = TRUE;
			OsDrawDemoInit();
			PostMessage(ghwndMain, WM_NULL, 0, 0);
		}
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hwnd, &ps);
			PaintParent(hdc);
			EndPaint(hwnd, &ps);
		}
		break;

	default:
		return DefWindowProc(hwnd, wm, wp, lp);
	}

	return 0;
}

